import asyncio
import aiohttp
import aiohttp.server
import time

from aiohttp.streams import EOF_MARKER

from . import errors
from .request import Request


__all__ = [
    'RESTRequestHandler',
    ]


class RESTRequestHandler(aiohttp.server.ServerHttpProtocol):

    def __init__(self,  server, *, hostname,
                 session_factory=None,
                 identity_policy=None, auth_policy=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.server = server
        self.hostname = hostname
        self.session_factory = session_factory
        self._identity_policy = identity_policy
        self._auth_policy = auth_policy

    @asyncio.coroutine
    def handle_request(self, message, payload):
        now = time.time()
        # self.log.debug("Start handle request %r at %d", message, now)

        try:
            if payload is not None:
                req_body = bytearray()
                while True:
                    chunk = yield from payload.readany()
                    req_body.extend(chunk)
                    if chunk is EOF_MARKER:
                        break
            else:
                req_body = None

            request = Request(self.hostname, message, req_body,
                              session_factory=self.session_factory,
                              loop=self._loop,
                              identity_policy=self._identity_policy,
                              auth_policy=self._auth_policy)

            body = yield from self.server.dispatch(request)
            bbody = body.encode('utf-8')
            resp_impl = aiohttp.Response(
                self.writer, request.response.status_code,
                http_version=message.version)

            yield from request._call_response_callbacks()

            resp_impl.add_header('Host', self.hostname)
            resp_impl.add_header('Content-Type', 'application/json')

            # content encoding
            accept_encoding = message.headers.get('ACCEPT-ENCODING',
                                                  '').lower()
            accept_encoding = set(map(str.strip, accept_encoding.split(',')))
            if 'deflate' in accept_encoding:
                resp_impl.add_header('Transfer-Encoding', 'chunked')
                resp_impl.add_header('Content-Encoding', 'deflate')
                resp_impl.add_compression_filter('deflate')
                resp_impl.add_chunking_filter(1025)
            elif 'gzip' in accept_encoding:
                resp_impl.add_header('Transfer-Encoding', 'chunked')
                resp_impl.add_header('Content-Encoding', 'gzip')
                resp_impl.add_compression_filter('gzip')
                resp_impl.add_chunking_filter(1025)
            else:
                resp_impl.add_header('Content-Length', str(len(bbody)))

            headers = request.response.headers.items()
            for key, val in headers:
                resp_impl.add_header(key, val)

            resp_impl.send_headers()
            resp_impl.write(bbody)
            yield from resp_impl.write_eof()
            if resp_impl.keep_alive():
                self.keep_alive(True)

            # self.log.debug("Fihish handle request %r at %d -> %s",
            #               message, time.time(), body)
            self.log_access(message, None, resp_impl, time.time() - now)
        except Exception:
            # self.log.exception("Cannot handle request %r", message)
            raise

    def handle_error(self, status=500, message=None, payload=None,
                     exc=None, headers=None):
        now = time.time()
        if isinstance(exc, errors.RESTError):
            if status == 500:
                self.log_exception("Error handling request")
            resp_impl = aiohttp.Response(self.writer, status, close=True)
            resp_impl.add_header('Host', self.hostname)
            yield from exc.write_response(resp_impl)
            self.log_access(message, None, resp_impl, time.time() - now)
            self.keep_alive(False)
        else:
            super().handle_error(status, message, payload,
                                 exc=exc, headers=headers)



import abc
import asyncio

# see http://plope.com/pyramid_auth_design_api_postmortem


class AbstractIdentityPolicy(metaclass=abc.ABCMeta):

    @asyncio.coroutine
    @abc.abstractmethod
    def identify(self, request):
        """ Return the claimed identity of the user associated request or
        ``None`` if no identity can be found associated with the request."""
        pass

    @asyncio.coroutine
    @abc.abstractmethod
    def remember(self, request, identity, **kw):
        """ Modify request.response which
        can be used to remember ``identity`` for a subsequent request.
        An individual identity policy and its consumers can decide on
        the composition and meaning of **kw."""
        pass

    @asyncio.coroutine
    @abc.abstractmethod
    def forget(self, request):
        """ Modify request.response which
        can be used to 'forget' the current identity
        on subsequent requests."""
        pass


class AbstractAuthorizationPolicy(metaclass=abc.ABCMeta):

    @asyncio.coroutine
    @abc.abstractmethod
    def permits(self, user_id, permission, context=None):
        """ Return True if the user_id is allowed the permission in the
        current context, else return False"""
        pass

    @asyncio.coroutine
    @abc.abstractmethod
    def authorized_user_id(self, identity):
        """ Return the user_id of the user identified by the identity
        or 'None' if no user exists related to the identity """
        pass


class CookieIdentityPolicy(AbstractIdentityPolicy):

    @asyncio.coroutine
    def identify(self, request):
        return request.cookies.get('user_id', None)

    @asyncio.coroutine
    def remember(self, request, identity, **kw):
        request.response.set_cookie('user_id', identity)

    @asyncio.coroutine
    def forget(self, request):
        request.response.del_cookie('user_id')



import asyncio
import json
import http.cookies

from urllib.parse import urlsplit, parse_qsl

from aiohttp.multidict import MultiDict, CIMultiDict, MultiDictProxy

from .errors import JsonLoadError, JsonDecodeError


__all__ = [
    'Request',
    'Response',
    ]


class Response:

    def __init__(self):
        self.headers = CIMultiDict()
        self._status_code = 200
        self._cookies = http.cookies.SimpleCookie()
        self._deleted_cookies = set()

    def _copy_cookies(self):
        for cookie in self._cookies.values():
            value = cookie.output(header='')[1:]
            self.headers.add('Set-Cookie', value)

    @property
    def cookies(self):
        return self._cookies

    def set_cookie(self, name, value, *, expires=None,
                   domain=None, max_age=None, path=None,
                   secure=None, httponly=None, version=None):
        """Set or update response cookie.

        Sets new cookie or updates existent with new value.
        Also updates only those params which are not None.
        """
        if name in self._deleted_cookies:
            self._deleted_cookies.remove(name)
            self._cookies.pop(name, None)

        self._cookies[name] = value
        c = self._cookies[name]
        if expires is not None:
            c['expires'] = expires
        if domain is not None:
            c['domain'] = domain
        if max_age is not None:
            c['max-age'] = max_age
        if path is not None:
            c['path'] = path
        if secure is not None:
            c['secure'] = secure
        if httponly is not None:
            c['httponly'] = httponly
        if version is not None:
            c['version'] = version

    def del_cookie(self, name, *, domain=None, path=None):
        """Delete cookie.

        Creates new empty expired cookie.
        """
        # TODO: do we need domain/path here?
        self._cookies.pop(name, None)
        self.set_cookie(name, '', max_age=0, domain=domain, path=path)
        self._deleted_cookies.add(name)

    @property
    def status_code(self):
        return self._status_code

    @status_code.setter
    def status_code(self, value):
        assert isinstance(value, int), "Status code must be int"
        self._status_code = value


class Request:

    def __init__(self, host, message, req_body, *,
                 session_factory=None, loop=None,
                 identity_policy=None, auth_policy=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        res = urlsplit(message.path)
        self._loop = loop
        self.version = message.version
        self.method = message.method.upper()
        self.host = message.headers.get('HOST', host)
        self.host_url = 'http://' + self.host
        self.path_qs = message.path
        self.path = res.path
        self.path_url = self.host_url + self.path
        self.url = self.host_url + self.path_qs
        self.query_string = res.query
        self.args = MultiDictProxy(MultiDict(parse_qsl(res.query)))
        self.headers = message.headers
        self.matchdict = {}
        self._request_body = req_body
        self._response = Response()
        self._session_factory = session_factory
        self._session_fut = None
        self._json_body = None
        self._cookies = None
        self._on_response = []
        self._identity_policy = identity_policy
        self._auth_policy = auth_policy

    @property
    def response(self):
        """Response object."""
        return self._response

    @property
    def session(self):
        if self._session_fut is None:
            self._session_fut = fut = asyncio.Future(loop=self._loop)
            if self._session_factory is not None:
                self._session_factory(self, fut)
            else:
                fut.set_result(None)
        return self._session_fut

    @property
    def json_body(self):
        if self._json_body is None:
            if self._request_body:
                # TODO: store generated exception and
                # don't try to parse json next time
                try:
                    decoded = self._request_body.decode('utf-8')
                    self._json_body = json.loads(decoded)
                except UnicodeDecodeError as exc:
                    raise JsonDecodeError(exc.encoding,
                                          exc.object,
                                          exc.start,
                                          exc.end,
                                          "JSON body is not utf-8 encoded",
                                          )
                except ValueError as exc:
                    raise JsonLoadError(
                        "JSON body can not be decoded", decoded)
            else:
                raise JsonLoadError("Request has no body")
        return self._json_body

    @property
    def cookies(self):
        """Return request cookies.

        A read-only dictionary-like object.
        """
        if self._cookies is None:
            raw = self.headers.get('COOKIE', '')
            parsed = http.cookies.SimpleCookie(raw)
            self._cookies = MultiDictProxy(MultiDict({key: val.value
                                                      for key, val in
                                                      parsed.items()}))
        return self._cookies

    def add_response_callback(self, callback, *args, **kwargs):
        """Add callback to be trigger when request is ready to be sent.
        """
        self._on_response.append((callback, args, kwargs))

    @asyncio.coroutine
    def _call_response_callbacks(self):
        callbacks = self._on_response[:]
        for callback, args, kwargs in callbacks:
            if asyncio.iscoroutinefunction(callback):
                yield from callback(self, *args, **kwargs)
            else:
                callback(self, *args, **kwargs)
        self.response._copy_cookies()

    @property
    def identity_policy(self):
        if not self._identity_policy:
            raise AttributeError('Identity policy not set')
        return self._identity_policy

    @property
    def auth_policy(self):
        if not self._auth_policy:
            raise AttributeError('Authorization policy not set')
        return self._auth_policy



import aiohttp
import asyncio
import json


class JsonLoadError(ValueError):
    pass


class JsonDecodeError(UnicodeDecodeError):
    pass


class RESTError(aiohttp.HttpProcessingError):
    """REST server error.

    Adds json_body to base aiohttp.HttpProcessingError
    """

    def __init__(self, code, message='', json_body={}, headers=None):
        super().__init__(code=code, message=message, headers=headers)
        if json_body is None:
            self.body = None
        else:
            body = {'error': json_body,
                    'error_reason': message,
                    'error_code': code}
            self.body = json.dumps(body).encode('utf-8')

    @asyncio.coroutine
    def write_response(self, response):
        if self.body is not None:
            response.add_headers(
                ('Content-Type', 'application/json; charset=utf-8'),
                ('Content-Length', str(len(self.body))),
                )
        else:
            response.add_headers(
                ('Content-Type', 'text/plain'),
                ('Content-Length', '0'),
                )
        if self.headers:
            response.add_headers(*self.headers)
        response.send_headers()

        if self.body is not None:
            response.write(self.body)
        yield from response.write_eof()


class HttpCorsOptions(RESTError):
    """Http exception to handle CORS preflight requests.

    Raised in RESTServer.dispatch.
    """

    def __init__(self, headers):
        super().__init__(200, json_body=None, headers=headers)
